# ==================================================================================================
# --- Imports
# ==================================================================================================

# Import third-party packages
import dash_mantine_components as dmc
from dash import dcc


# ==================================================================================================
# --- 3D separation Layout
# ==================================================================================================
def return_3D_separation_layout():
    """
    Returns the layout for the 3D beam separation dashboard.

    Returns:
        separation_layout : dash_bootstrap_components.Center
            Layout for the 3D beam separation dashboard.
    """
    separation_layout = (
        dmc.Center(
            dmc.Stack(
                children=[
                    dmc.Center(
                        children=[
                            dmc.Group(
                                children=[
                                    dmc.Group(
                                        children=[
                                            dmc.Text("Beam-beam: "),
                                            dmc.ChipGroup(
                                                [
                                                    dmc.Chip(
                                                        x,
                                                        value=x,
                                                        variant="outline",
                                                        color="cyan",
                                                    )
                                                    for x in ["On", "Off"]
                                                ],
                                                id="chips-sep-bb-3D",
                                                value="Off",
                                                mb=0,
                                            ),
                                        ],
                                        pt=5,
                                    ),
                                    dmc.Space(w="xl"),
                                    dmc.Group(
                                        children=[
                                            dmc.Text("IP: "),
                                            dmc.ChipGroup(
                                                [
                                                    dmc.Chip(
                                                        x,
                                                        value=x,
                                                        variant="outline",
                                                        color="cyan",
                                                    )
                                                    for x in ["IP 1", "IP 2", "IP 5", "IP 8"]
                                                ],
                                                id="chips-ip-3D",
                                                value="IP 1",
                                                mb=0,
                                            ),
                                        ],
                                        pt=5,
                                    ),
                                ],
                            ),
                        ],
                    ),
                    dcc.Loading(
                        dcc.Graph(
                            id="beam-separation-3D",
                            mathjax=True,
                            config={
                                "displayModeBar": False,
                                "scrollZoom": True,
                                "responsive": True,
                                "displaylogo": False,
                            },
                            style={"height": "90vh", "width": "100%", "margin": "auto"},
                        ),
                        type="circle",
                        color="cyan",
                        style={"height": "100%", "width": "100%", "margin": "auto"},
                        parent_style={
                            "height": "100%",
                            "width": "100%",
                            "margin": "auto",
                        },
                    ),
                ],
                style={"width": "100%", "margin": "auto"},
            )
        ),
    )
    return separation_layout



"""Module containing the callbacks for the app."""

# ==================================================================================================
# --- Imports
# ==================================================================================================

# Import third-party packages
import dash_mantine_components as dmc
import plotly.graph_objects as go
from dash import Input, Output, no_update

# Import initialization and plotting functions
from .backend import plot

# Import layout functions
from .layout import (
    return_3D_separation_layout,
    return_configuration_layout,
    return_filling_scheme_layout,
    return_footprint_layout,
    return_knobs_layout,
    return_optics_layout,
    return_sanity_layout,
    return_separation_layout,
    return_survey_layout,
    return_tables_layout,
)


# ==================================================================================================
# --- Functions used by callbacks
# ==================================================================================================
def return_tabs_sanity(dic_with_bb, dic_without_bb):
    sanity_after_beam_beam = return_sanity_layout(
        dic_with_bb["dic_tw_b1"],
        dic_with_bb["dic_tw_b2"],
        dic_with_bb["l_lumi"],
        dic_with_bb["array_b1"],
        dic_with_bb["array_b2"],
        dic_with_bb["polarity_alice"],
        dic_with_bb["polarity_lhcb"],
        dic_with_bb["energy"],
    )
    sanity_before_beam_beam = return_sanity_layout(
        dic_without_bb["dic_tw_b1"],
        dic_without_bb["dic_tw_b2"],
        dic_without_bb["l_lumi"],
        dic_without_bb["array_b1"],
        dic_without_bb["array_b2"],
        dic_without_bb["polarity_alice"],
        dic_without_bb["polarity_lhcb"],
        dic_with_bb["energy"],
    )
    tabs_sanity = dmc.Tabs(
        [
            dmc.TabsList(
                [
                    dmc.Tab(
                        "Before beam-beam",
                        value="sanity-before-beam-beam",
                        style={"font-size": "1.1rem"},
                    ),
                    dmc.Tab(
                        "After beam beam",
                        value="sanity-after-beam-beam",
                        style={"font-size": "1.1rem"},
                    ),
                ],
                position="center",
            ),
            dmc.TabsPanel(sanity_before_beam_beam, value="sanity-before-beam-beam"),
            dmc.TabsPanel(sanity_after_beam_beam, value="sanity-after-beam-beam"),
        ],
        color="cyan",
        value="sanity-after-beam-beam",
    )

    return tabs_sanity


# ==================================================================================================
# --- App callbacks
# ==================================================================================================


def all_callbacks(app, dic_with_bb, dic_without_bb, path_collider):
    @app.callback(Output("placeholder-tabs", "children"), Input("tab-titles", "value"))
    def select_tab(value):
        match value:
            case "display-configuration":
                return return_configuration_layout(dic_with_bb["configuration_str"], path_collider)
            case "display-knobs":
                return return_knobs_layout(dic_with_bb["dic_knob_str"])
            case "display-twiss":
                return return_tables_layout()
            case "display-scheme":
                return return_filling_scheme_layout()
            case "display-separation":
                return return_separation_layout()
            case "display-3D-separation":
                return return_3D_separation_layout()
            case "display-footprint":
                return return_footprint_layout()
            case "display-sanity":
                return return_tabs_sanity(dic_with_bb, dic_without_bb)
            case "display-optics":
                return return_optics_layout(dic_with_bb)
            case "display-survey":
                return return_survey_layout()
            case _:
                return return_configuration_layout(dic_with_bb["configuration_str"], path_collider)

    @app.callback(
        Output("prism-knob_info", "children"),
        Input("select-knob_info", "value"),
    )
    def select_knob_info(knob):
        if knob is not None:
            return dic_with_bb["dic_knob_str"][knob]
        else:
            return no_update

    @app.callback(
        Output("placeholder-data-table", "children"),
        Input("segmented-data-table", "value"),
    )
    def select_data_table(value):
        match value:
            case "Twiss table beam 1":
                return dic_with_bb["table_tw_b1"]
            case "Survey table beam 1":
                return dic_with_bb["table_sv_b1"]
            case "Twiss table beam 2":
                return dic_with_bb["table_tw_b2"]
            case "Survey table beam 2":
                return dic_with_bb["table_sv_b2"]
            case _:
                return dic_with_bb["table_tw_b1"]

    @app.callback(
        Output("LHC-layout", "figure"),
        Input("chips-ip", "value"),
    )
    def update_graph_LHC_layout(l_values):
        l_indices_to_keep = []
        for val in l_values:
            str_ind_1, str_ind_2 = val.split("-")
            # Get indices of elements to keep (# ! implemented only for beam 1)
            l_indices_to_keep.extend(
                plot.get_indices_of_interest(
                    dic_with_bb["df_tw_b1"], "ip" + str_ind_1, "ip" + str_ind_2
                )
            )

        fig = plot.return_plot_lattice_with_tracking(
            dic_with_bb["df_sv_b1"],
            dic_with_bb["df_elements_corrected"],
            dic_with_bb["df_tw_b1"],
            df_sv_2=dic_with_bb["df_sv_b2"],
            df_tw_2=dic_with_bb["df_tw_b2"],
            l_indices_to_keep=l_indices_to_keep,
        )

        return fig

    @app.callback(
        Output("filling-scheme-graph", "figure"),
        Output("filling-scheme-graph", "style"),
        Output("filling-scheme-alert", "style"),
        Input("tab-titles", "value"),
    )
    def update_graph_filling(value):
        if value == "display-scheme":
            if dic_with_bb["array_b1"] is not None:
                return (
                    plot.return_plot_filling_scheme(
                        dic_with_bb["array_b1"],
                        dic_with_bb["array_b2"],
                        dic_with_bb["i_bunch_b1"],
                        dic_with_bb["i_bunch_b2"],
                        dic_with_bb["bbs"],
                    ),
                    {"height": "90vh", "width": "100%", "margin": "auto"},
                    {"margin": "auto", "display": "none"},
                )

            else:
                return (
                    go.Figure(),
                    {
                        "height": "90vh",
                        "width": "100%",
                        "margin": "auto",
                        "display": "none",
                    },
                    {"margin": "auto"},
                )

        else:
            return no_update

    @app.callback(
        Output("LHC-2D-near-IP", "figure"),
        Input("tab-titles", "value"),
        Input("vertical-zoom-optics", "value"),
    )
    def update_graph_optics(tab_value, zoom_value):
        if tab_value == "display-optics":
            fig = plot.return_plot_optics(
                dic_with_bb["df_tw_b1"],
                dic_with_bb["df_tw_b2"],
                dic_with_bb["df_sv_b1"],
                dic_with_bb["df_elements_corrected"],
            )

            factor = 2**-zoom_value
            fig.update_yaxes(
                title_text=r"$\beta_{x,y}[m]$",
                range=[0, 10000 * factor * 2],
                row=2,
                col=1,
            )
            fig.update_yaxes(
                title_text=r"(Closed orbit)$_{x,y}[m]$",
                range=[-0.03 * factor, 0.03 * factor],
                row=3,
                col=1,
            )
            fig.update_yaxes(
                title_text=r"$D_{x,y}[m]$",
                range=[-3 * factor, 3 * factor],
                row=4,
                col=1,
            )

            return fig
        else:
            return no_update

    @app.callback(
        Output("beam-separation", "figure"),
        Input("chips-sep", "value"),
        Input("chips-sep-bb", "value"),
    )
    def update_graph_separation(value, bb):
        if bb == "On":
            dic = dic_with_bb
        elif bb == "Off":
            dic = dic_without_bb
        else:
            raise ValueError("bb should be either On or Off")

        if value == "v" or value == "h":
            fig = plot.return_plot_separation(
                dic["dic_separation_ip"], "x" if value == "h" else "y"
            )
        elif value == "||v+h||":
            fig = plot.return_plot_separation(dic["dic_separation_ip"], "xy")
        else:
            raise ValueError("value should be either v, h or ||v+h||")
        return fig

    @app.callback(
        Output("beam-separation-3D", "figure"),
        Input("chips-sep-bb-3D", "value"),
        Input("chips-ip-3D", "value"),
    )
    def update_graph_separation_3D(bb, ip):
        if bb == "On":
            dic = dic_with_bb
        elif bb == "Off":
            dic = dic_without_bb
        else:
            raise ValueError("bb should be either On or Off")

        fig = plot.return_plot_separation_3D(dic["dic_position_ip"], ip.lower().replace(" ", ""))

        return fig

    @app.callback(
        Output("footprint-without-bb-b1", "figure"),
        Output("footprint-without-bb-b2", "figure"),
        Output("footprint-with-bb-b1", "figure"),
        Output("footprint-with-bb-b2", "figure"),
        Input("tab-titles", "value"),
    )
    def update_graph_footprint(value):
        if value == "display-footprint":
            if dic_without_bb["i_bunch_b1"] is not None:
                title_without_bb_b1 = (
                    "Tune footprint without beam-beam effects for beam 1 and bunch "
                    + str(dic_without_bb["i_bunch_b1"])
                )
                title_without_bb_b2 = (
                    "Tune footprint without beam-beam effects for beam 2 and bunch "
                    + str(dic_without_bb["i_bunch_b2"])
                )
                title_with_bb_b1 = (
                    "Tune footprint with beam-beam effects for beam 1 and bunch "
                    + str(dic_with_bb["i_bunch_b1"])
                )
                title_with_bb_b2 = (
                    "Tune footprint with beam-beam effects for beam 2 and bunch "
                    + str(dic_with_bb["i_bunch_b2"])
                )
            else:
                title_without_bb_b1 = (
                    "Tune footprint without beam-beam effects for beam 1 (bunch number unknown)"
                )
                title_without_bb_b2 = (
                    "Tune footprint without beam-beam effects for beam 2 (bunch number unknown)"
                )
                title_with_bb_b1 = (
                    "Tune footprint with beam-beam effects for beam 1 (bunch number unknown)"
                )
                title_with_bb_b2 = (
                    "Tune footprint with beam-beam effects for beam 2 (bunch number unknown)"
                )

            return [
                plot.return_plot_footprint(
                    dic_without_bb["footprint_b1"],
                    dic_without_bb["dic_tw_b1"]["qx"],
                    dic_without_bb["dic_tw_b1"]["qy"],
                    title=title_without_bb_b1,
                ),
                plot.return_plot_footprint(
                    dic_without_bb["footprint_b2"],
                    dic_without_bb["dic_tw_b2"]["qx"],
                    dic_without_bb["dic_tw_b2"]["qy"],
                    title=title_without_bb_b2,
                ),
                plot.return_plot_footprint(
                    dic_with_bb["footprint_b1"],
                    dic_without_bb["dic_tw_b1"]["qx"],
                    dic_without_bb["dic_tw_b1"]["qy"],
                    title=title_with_bb_b1,
                ),
                plot.return_plot_footprint(
                    dic_with_bb["footprint_b2"],
                    dic_without_bb["dic_tw_b2"]["qx"],
                    dic_without_bb["dic_tw_b2"]["qy"],
                    title=title_with_bb_b2,
                ),
            ]
        else:
            return no_update



# ==================================================================================================
# --- Imports
# ==================================================================================================

# Import third-party packages
import dash_mantine_components as dmc
from dash import dcc


# ==================================================================================================
# --- Filling scheme Layout
# ==================================================================================================
def return_filling_scheme_layout():
    """
    Returns the layout for the filling scheme analysis dashboard page.

    Returns:
        scheme_layout : dash.development.base_component.Component
            The layout for the filling scheme analysis dashboard page.
    """
    scheme_layout = dmc.Stack(
        children=[
            dmc.Alert(
                (
                    "Filling scheme analysis not available as no configuration was provided when"
                    " building the dashboard"
                ),
                title="No configuration provided!",
                id="filling-scheme-alert",
                style={"margin": "auto", "display": "none"},
            ),
            dcc.Loading(
                dcc.Graph(
                    id="filling-scheme-graph",
                    mathjax=True,
                    config={
                        "displayModeBar": False,
                        "scrollZoom": True,
                        "responsive": True,
                        "displaylogo": False,
                    },
                    style={"height": "90vh", "width": "100%", "margin": "auto"},
                ),
                type="circle",
                color="cyan",
            ),
        ]
    )
    return scheme_layout



# ==================================================================================================
# --- Imports
# ==================================================================================================
import dash_mantine_components as dmc


# ==================================================================================================
# --- Configuration layout
# ==================================================================================================
def return_configuration_layout(configuration_str, path_collider):
    """
    Returns a Dash layout for displaying a YAML configuration file.

    Args:
        configuration_str (str): The YAML configuration file as a string.
        path_collider (str): The path to the collider file (located in the same directory as the
            configuration).

    Returns:
        dash.development.base_component.Component: A Dash layout component.
    """
    if configuration_str is not None:
        configuration_layout = dmc.Stack(
            children=[
                dmc.Center(dmc.Text("Configuration path: " + path_collider)),
                dmc.Center(
                    children=[
                        dmc.Prism(
                            language="yaml",
                            children=configuration_str,
                            style={
                                "height": "90vh",
                                "overflowY": "auto",
                                "width": "80%",
                            },
                        ),
                    ],
                ),
            ],
        )

    else:
        configuration_layout = dmc.Center(
            dmc.Alert(
                "Configuration can't be displayed as no configuration file was provided.",
                title="No configuration !",
                style={"margin": "auto"},
            ),
            style={"width": "70%", "margin": "auto"},
        )

    return configuration_layout



class dotdict(dict):
    '''
    A dict with dot access and autocompletion.
    
    HINT: use as keys string starting with 'at' or 'b' to ease the autocompletion from the drop down menu. 

    The idea and most of the code was taken from 
    http://stackoverflow.com/a/23689767,
    http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named/
    http://stackoverflow.com/questions/2390827/how-to-properly-subclass-dict-and-override-get-set
    
    ===Example===     

    a=dotdict()
    for i in {'B1','B2'}:
        a[i]=dotdict()
        for j in range(100):
            a[i]['b'+str(j)]=np.random.randn(10)
    '''
    
    def __init__(self,*a,**kw):
        dict.__init__(self)
        self.update(*a, **kw)
        self.__dict__ = self
    
    def __setattr__(self, key, value):
        if key in dict.__dict__:
            raise AttributeError('This key is reserved for the dict methods.')
        dict.__setattr__(self, key, value)
    
    def __setitem__(self, key, value):
        if key in dict.__dict__:
            raise AttributeError('This key is reserved for the dict methods.')
        dict.__setitem__(self, key, value)
        
    def update(self, *args, **kwargs):
        for k, v in list(dict(*args, **kwargs).items()):
            self[k] = v
        
    def __getstate__(self):
        return self
 
    def __setstate__(self, state):
        self.update(state)
        self.__dict__ = self



"""Thid module initializes all global variables from a collider json, potentially embedding a 
configuration file.
"""

# ==================================================================================================
# --- Imports
# ==================================================================================================

# Import from standard library
import io
import logging
import os
import pickle
from contextlib import redirect_stdout
from importlib.resources import files

# Third-party packages
import numpy as np
import pandas as pd
import xtrack as xt

# Package to check collider observables
from collider_check import ColliderCheck

# Dash imports
from dash import dash_table
from dash.dash_table.Format import Format, Scheme

# Package to compute beam-beam schedule
from .fillingpatterns import FillingPattern

# ==================================================================================================
# --- Functions initialize all global variables
# ==================================================================================================


def init_from_collider(
    path_collider, path_scheme=None, force_reload=False, ignore_footprint=False, simplify_tw=True
):
    """
    Initializes a collider from a JSON file and computes global variables from collider checks.

    Args:
        path_collider (str): Path to the JSON file containing the collider definition.
        force_reload (bool, optional): If False, tries to to load the global variables
            from a pickle file instead of computing them from scratch. Otherwise, force the
            (re-)computation of the global variables. Defaults to False.
        ignore_footprint (bool, optional): If True, does not compute the footprints to gain loading
            time. Defaults to False.
        simplify_tw (bool, optional): If True, simplifies the Twiss and Survey dataframes by
            removing duplicated elements to speed up computations. Defaults to True.

    Returns:
        tuple: A tuple containing two dictionaries of global variables computed from collider
            checks, one for the collider with beam-beam interactions and one for the collider
            without beam-beam interactions, and the path to the pickle file used to store the global
            variables.
    """

    # Path to the pickle dictionnaries (for loading and saving)
    temp_path = str(files("collider_dashboard")) + "/temp/"
    path_pickle = temp_path + path_collider.replace("/", "_") + "t_dic_var.pkl"

    # Check that the pickle file exists
    if not os.path.isfile(path_pickle) or force_reload:
        print(
            "No data has been recorded for this collider, or force_reload is True. Recomputing the"
            " dashboard collider data now."
        )

        logging.info("Building collider.")
        # Rebuild collider
        collider = xt.Multiline.from_json(path_collider)

        # Make a copy of the collider to load without bb after
        collider_without_bb = xt.Multiline.from_dict(collider.to_dict())

        # Load collider with bb
        collider.build_trackers()

        # Build collider before bb
        collider_without_bb.build_trackers()
        collider_without_bb.vars["beambeam_scale"] = 0

        # Check configuration
        if collider.metadata is not None and collider.metadata != {}:
            logging.info("The collider file contains metadata. Using it.")
        else:
            logging.warning("The collider file does not contain metadata. Using default values.")

        # Compute collider checks
        logging.info("Computing collider checks.")
        collider_check_with_bb = ColliderCheck(collider, path_filling_scheme=path_scheme)
        collider_check_without_bb = ColliderCheck(
            collider_without_bb, path_filling_scheme=path_scheme
        )

        # Compute global variables
        dic_without_bb, dic_with_bb = compute_global_variables_from_collider_checks(
            collider_check_with_bb,
            collider_check_without_bb,
            path_pickle=path_pickle,
            ignore_footprint=ignore_footprint,
            simplify_tw=simplify_tw,
        )

    else:
        print("Some collider data already exists for this path, loading it.")
        logging.info("Returning global variables from pickle file.")
        with open(path_pickle, "rb") as f:
            dic_without_bb, dic_with_bb = pickle.load(f)

    return dic_without_bb, dic_with_bb


def compute_global_variables_from_collider_checks(
    collider_check_after_beam_beam,
    collider_check_without_beam_beam,
    path_pickle=None,
    ignore_footprint=False,
    simplify_tw=True,
):
    """
    Computes global variables with and without beam-beam interaction.

    Args:
        collider_check_after_beam_beam (ColliderCheck): Collider check object including beam-beam
            interaction.
        collider_check_without_beam_beam (ColliderCheck): Collider check object without beam-beam
            interaction.
        path_pickle (str, optional): Path to the pickle file to dump the dictionaries. Defaults to
            None.
        ignore_footprint (bool, optional): If True, does not compute the footprints to gain loading
            time. Defaults to False.
        simplify_tw (bool, optional): If True, simplifies the Twiss and Survey dataframes by
            removing duplicated elements to speed up computations. Defaults to True.


    Returns:
        tuple: A tuple containing two dictionaries, one for global variables with beam-beam
            interaction and one for global variables without beam-beam interaction.
    """
    # Get the global variables before and after the beam-beam
    logging.info("Computing global variables with beam beam.")
    dic_with_bb = initialize_global_variables(
        collider_check_after_beam_beam,
        compute_footprint=not ignore_footprint,
        simplify_tw=simplify_tw,
    )
    logging.info("Computing global variables without beam beam.")
    dic_without_bb = initialize_global_variables(
        collider_check_without_beam_beam,
        compute_footprint=not ignore_footprint,
        simplify_tw=simplify_tw,
    )

    if path_pickle is not None:
        # Dump the dictionnaries in a pickle file, creating the directory if it does not exist
        if not os.path.isdir(os.path.dirname(path_pickle)):
            os.makedirs(os.path.dirname(path_pickle))
        logging.info("Dumping global variables into a pickle file.")
        with open(path_pickle, "wb") as f:
            pickle.dump((dic_without_bb, dic_with_bb), f)

    return dic_without_bb, dic_with_bb


def initialize_global_variables(collider_check, compute_footprint=True, simplify_tw=True):
    """
    Initializes global variables used in the simulation dashboard.

    Parameters:
    -----------
    collider_check : ColliderCheck
        An instance of the ColliderCheck class containing the collider configuration.
    compute_footprint : bool, optional
        Whether to compute the collider footprint or not. Default is True.
    simplify_tw (bool, optional): If True, simplifies the Twiss and Survey dataframes by removing
        duplicated elements to speed up computations. Defaults to True.

    Returns:
    --------
    dic_global_var : dict
        A dictionary containing the initialized global variables.
    """

    if collider_check.configuration is not None:
        # Get luminosity at each IP
        logging.info("Computing luminosity at each IP.")
        l_lumi = [collider_check.return_luminosity(IP=x) for x in [1, 2, 5, 8]]

        # Get the beams schemes
        array_b1 = collider_check.array_b1
        array_b2 = collider_check.array_b2

        # Get the bunches selected for tracking
        i_bunch_b1 = collider_check.i_bunch_b1
        i_bunch_b2 = collider_check.i_bunch_b2

        # Get emittances
        nemitt_x = collider_check.nemitt_x
        nemitt_y = collider_check.nemitt_y

        # Get energy
        energy = collider_check.energy

        # Get the beam-beam schedule
        logging.info("Computing beam-beam schedule.")
        patt = FillingPattern.from_json(collider_check.path_filling_scheme)
        patt.compute_beam_beam_schedule(n_lr_per_side=26)
        bbs = patt.b1.bb_schedule

        # Get polarity Alice and LHCb
        polarity_alice, polarity_lhcb = collider_check.return_polarity_ip_2_8()

        # Get configuration
        configuration_str = collider_check.configuration_str

    else:
        l_lumi = None
        array_b1 = None
        array_b2 = None
        i_bunch_b1 = None
        i_bunch_b2 = None
        bbs = None
        polarity_alice = None
        polarity_lhcb = None
        configuration_str = None
        energy = None

        # Get emittance for the computation of the normalized separation
        logging.warning("No configuration file provided, using default values for emittances.")
        nemitt_x = 2.2e-6
        nemitt_y = 2.2e-6

    # Get elements of the line (only done for b1, should be identical for b2)
    logging.info("Getting beam-beam elements for plotting.")
    df_elements = return_dataframe_elements_from_line(collider_check.collider.lhcb1)

    # Get twiss and survey for both lines
    logging.info("Computing Twiss and Survey.")
    tw_b1, df_sv_b1, df_tw_b1 = (
        collider_check.tw_b1,
        collider_check.df_sv_b1,
        collider_check.df_tw_b1,
    )

    tw_b2, sv_b2 = collider_check.tw_b2.reverse(), collider_check.sv_b2.reverse()
    df_tw_b2, df_sv_b2 = tw_b2.to_pandas(), sv_b2.to_pandas()

    # Correct df elements for thin lens approximation
    logging.info("Correcting beam-beam elements for thin lens approximation.")
    df_elements_corrected = return_dataframe_corrected_for_thin_lens_approx(df_elements, df_tw_b1)

    # Get corresponding data tables
    logging.info("Get Twiss and survey datatables.")
    if simplify_tw:
        logging.info("Simplifying Twiss and survey datatables as requested.")

    table_sv_b1, df_sv_b1 = return_data_table(
        df_sv_b1, "id-df-sv-b1-after-bb", twiss=False, simplify_tw=simplify_tw
    )
    table_tw_b1, df_tw_b1 = return_data_table(
        df_tw_b1, "id-df-tw-b1-after-bb", twiss=True, simplify_tw=simplify_tw
    )
    table_sv_b2, df_sv_b2 = return_data_table(
        df_sv_b2, "id-df-sv-b2-after-bb", twiss=False, simplify_tw=simplify_tw
    )
    table_tw_b2, df_tw_b2 = return_data_table(
        df_tw_b2, "id-df-tw-b2-after-bb", twiss=True, simplify_tw=simplify_tw
    )

    # Get the twiss dictionary (tune, chroma, etc + twiss at IPs)
    logging.info("Get Twiss dictionary.")
    dic_tw_b1 = return_twiss_dic(tw_b1)
    dic_tw_b2 = return_twiss_dic(tw_b2)

    # Get the dictionary to plot separation
    logging.info("Computing separation variables")
    dic_separation_ip = {
        f"ip{ip}": collider_check.compute_separation_variables(ip=f"ip{ip}") for ip in [1, 2, 5, 8]
    }
    dic_position_ip = collider_check.return_dic_position_all_ips()

    # Convert the twiss variables in dic_separation_ip to pandas dataframe so that it can be saved in a pickle file
    logging.info("Converting twiss and survey objects to pandas dataframes.")
    for ip in [1, 2, 5, 8]:
        for variable_to_convert in [
            "twiss_filtered",
            "survey_filtered",
            "s",
            "dx_meter",
            "dy_meter",
            "dx_sig",
            "dy_sig",
        ]:
            if variable_to_convert == "twiss_filtered" or variable_to_convert == "survey_filtered":
                dic_separation_ip[f"ip{ip}"][variable_to_convert]["b1"] = dic_separation_ip[
                    f"ip{ip}"
                ][variable_to_convert]["b1"].to_pandas()
                dic_separation_ip[f"ip{ip}"][variable_to_convert]["b2"] = dic_separation_ip[
                    f"ip{ip}"
                ][variable_to_convert]["b2"].to_pandas()
            else:
                dic_separation_ip[f"ip{ip}"][variable_to_convert] = np.array(
                    dic_separation_ip[f"ip{ip}"][variable_to_convert], dtype=np.float64
                )

    # Get the footprint only if bb is on
    array_qx1 = np.array([])
    array_qy1 = np.array([])
    array_qx2 = np.array([])
    array_qy2 = np.array([])
    if compute_footprint:
        logging.info("Computing footprints.")
        try:
            array_qx1, array_qy1 = return_footprint(
                collider_check.collider, nemitt_x, beam="lhcb1", n_turns=2000
            )
            array_qx2, array_qy2 = return_footprint(
                collider_check.collider, nemitt_x, beam="lhcb2", n_turns=2000
            )
        except KeyError:
            logging.warning(
                "Could not compute footprint as 'beambeam_scale' is probably non existent."
            )
        except AssertionError:
            logging.warning(AssertionError)

    # Get knobs
    logging.info("Getting knobs.")
    dic_knob_str = compute_knob_str(collider_check)

    # Store everything in a dictionary
    dic_global_var = {
        "l_lumi": l_lumi,
        "dic_tw_b1": dic_tw_b1,
        "dic_tw_b2": dic_tw_b2,
        "dic_separation_ip": dic_separation_ip,
        "dic_position_ip": dic_position_ip,
        "df_sv_b1": df_sv_b1,
        "df_sv_b2": df_sv_b2,
        "df_tw_b1": df_tw_b1,
        "df_tw_b2": df_tw_b2,
        "df_elements_corrected": df_elements_corrected,
        "table_sv_b1": table_sv_b1,
        "table_tw_b1": table_tw_b1,
        "table_sv_b2": table_sv_b2,
        "table_tw_b2": table_tw_b2,
        "array_b1": array_b1,
        "array_b2": array_b2,
        "i_bunch_b1": i_bunch_b1,
        "i_bunch_b2": i_bunch_b2,
        "bbs": bbs,
        "footprint_b1": (array_qx1, array_qy1),
        "footprint_b2": (array_qx2, array_qy2),
        "polarity_alice": polarity_alice,
        "polarity_lhcb": polarity_lhcb,
        "energy": energy,
        "configuration_str": configuration_str,
        "dic_knob_str": dic_knob_str,
    }

    return dic_global_var


# ==================================================================================================
# --- Functions to load dashboard variables
# ==================================================================================================
def return_dataframe_elements_from_line(line):
    """
    Returns a pandas DataFrame containing the elements of a given line object.

    Args:
    line (Line): A Line object containing elements to be extracted into a DataFrame.

    Returns:
    df_elements (pandas.DataFrame): A DataFrame containing the elements of the given line object.
    """
    df_elements = pd.DataFrame([x.to_dict() for x in line.elements])
    return df_elements


def return_dataframe_corrected_for_thin_lens_approx(df_elements, df_tw):
    """
    Corrects the dataframe of elements for thin lens approximation.

    Parameters:
    -----------
    df_elements : pandas.DataFrame
        The dataframe of elements to be corrected.
    df_tw : pandas.DataFrame
        The corresponding Twiss from which the corrion is computed.

    Returns:
    --------
    pandas.DataFrame
        The corrected dataframe of elements.
    """

    df_elements_corrected = df_elements.copy(deep=True)

    # Get duplicated elements, according to regex, whose name:
    # - does not contain the words "entry" or "exit" or ends with "f".
    # - does not contain a period before the end of the string.
    # - contain exactly two consecutive periods.
    df_tw_duplicated_elements = df_tw[
        df_tw.name.str.contains("^(?!.*(?:entry|exit|[^f]*f[^.]*$)).*\.{2}.*", regex=True)
    ]

    # Get original elements
    df_tw_original_elements = df_tw[
        df_tw.name.isin(df_tw_duplicated_elements.name.str.split("..", regex=False).str[0])
    ]

    # Add all thin lenses (length + strength)
    for i, row in df_tw_duplicated_elements.iterrows():
        # Correct for thin lens approximation and weird duplicates
        name = row["name"].split("..")[0]
        try:
            index = df_tw_original_elements[df_tw_original_elements.name == name].index[0]
        except IndexError:
            print(f"IndexError trying to correct slicing for {name}")
            continue

        # Add length
        if np.isnan(df_elements_corrected.loc[index]["length"]):
            df_elements_corrected.at[index, "length"] = 0.0
        df_elements_corrected.at[index, "length"] += df_elements.loc[i]["length"]

        # Add strength
        if np.isnan(df_elements_corrected.loc[index]["knl"]).all():
            df_elements_corrected.at[index, "knl"] = (
                np.array([0.0] * df_elements.loc[i]["knl"].shape[0], dtype=np.float64)
                if not isinstance(df_elements.loc[i]["knl"], float)
                else 0.0
            )

        df_elements_corrected.at[index, "knl"] = (
            df_elements_corrected.loc[index, "knl"] + np.array(df_elements.loc[i]["knl"])
            if not isinstance(df_elements.loc[i]["knl"], float)
            else df_elements.loc[i]["knl"]
        )

        # Replace order
        df_elements_corrected.at[index, "_order"] = df_elements.loc[i]["_order"]

    # Drop all duplicate rows
    df_elements_corrected.drop(index=df_tw_duplicated_elements.index, inplace=True)

    return df_elements_corrected


def return_twiss_dic(tw):
    """
    Returns a dictionary containing important Twiss parameters.

    Parameters:
    tw (Twiss): A Twiss object for the current line.

    Returns:
    dict: A dictionary containing the following keys:
        - qx (float): Horizontal tune.
        - qy (float): Vertical tune.
        - dqx (float): Horizontal chromaticity.
        - dqy (float): Vertical chromaticity.
        - c_minus (float): Linear coupling.
        - momentum_compaction_factor (float): Momentum compaction factor.
        - T_rev0 (float): Revolution period.
        - ip1 (numpy.ndarray): Momentum, position and beta functions at IP1.
        - ip2 (numpy.ndarray): Momentum, position and beta functions at IP2.
        - ip5 (numpy.ndarray): Momentum, position and beta functions at IP5.
        - ip8 (numpy.ndarray): Momentum, position and beta functions at IP8.
    """
    # Init empty dic
    dic_tw = {}

    # Load main observables
    dic_tw["qx"] = tw["qx"]
    dic_tw["qy"] = tw["qy"]
    dic_tw["dqx"] = tw["dqx"]
    dic_tw["dqy"] = tw["dqy"]
    dic_tw["c_minus"] = tw["c_minus"]
    dic_tw["momentum_compaction_factor"] = tw["momentum_compaction_factor"]
    dic_tw["T_rev0"] = tw["T_rev0"]

    # Load observables at IPs
    for ip in [1, 2, 5, 8]:
        dic_tw["ip" + str(ip)] = (
            tw.rows[f"ip{ip}"]
            .cols["s", "x", "px", "y", "py", "betx", "bety"]
            .to_pandas()
            .to_numpy()
            .squeeze()
        )

    return dic_tw


# ==================================================================================================
# --- Functions to build data tables
# ==================================================================================================
def return_data_table(df, id_table, twiss=True, simplify_tw=True):
    """
    Returns a (stylized) Dash DataTable object containing the twiss/survey data of a given line
        (through the pandas dataframe input).

    Args:
        df (pandas.DataFrame): The DataFrame containing the twiss/survey data to be displayed in
            the table.
        id_table (str): The ID to be assigned to the DataTable object (for Dash callback).
        twiss (bool, optional): Whether or not the DataFrame contains twiss data. Defaults to True.
            If False, it is assumed to contain survey data.
        simplify_tw (bool, optional): If True, simplifies the Twiss and Survey dataframes by
            removing duplicated elements to speed up computations. Defaults to True.

    Returns:
        dash_table.DataTable: The DataTable object populated with the data from the input DataFrame.
    """
    if twiss:
        df = df.drop(["W_matrix"], axis=1)
        idx_column_name = 0
    else:
        idx_column_name = 6

    # Change order of columns such that name is first
    df = df[["name"] + [col for col in df.columns if col != "name"]]

    # Simplify the dataframe removing all duplicated elements, entry and exit
    if simplify_tw:
        df = df[df["name"].str.contains(r"^(?:(?!\.\.|entry|exit).)*$", regex=True)]

    table = (
        dash_table.DataTable(
            id=id_table,
            columns=[
                (
                    {
                        "name": i,
                        "id": i,
                        "deletable": False,
                        "type": "numeric",
                        "format": Format(precision=6, scheme=Scheme.decimal_or_exponent),
                    }
                    if idx != idx_column_name
                    else {"name": i, "id": i, "deletable": False}
                )
                for idx, i in enumerate(df.columns)
            ],
            data=df.to_dict("records"),
            editable=False,
            filter_action="native",
            sort_action="native",
            sort_mode="multi",
            row_selectable=False,
            row_deletable=False,
            virtualization=False,
            page_size=25,
            page_current=0,
            # page_action="custom",
            # filter_action="custom",
            # filter_query="",
            # sort_action="custom",
            # sort_mode="multi",
            # sort_by=[],
            style_table={
                # "height": "100%",
                "maxHeight": "75vh",
                "margin-x": "auto",
                "margin-top": "20px",
                "overflowY": "auto",
                "overflowX": "auto",
                "minWidth": "98%",
                "padding": "1em",
            },
            style_header={
                "backgroundColor": "rgb(30, 30, 30)",
                "color": "white",
                "padding": "1em",
            },
            style_data={"backgroundColor": "rgb(50, 50, 50)", "color": "white"},
            style_filter={"backgroundColor": "rgb(70, 70, 70)"},  # , "color": "white"},
            style_cell={"font-family": "sans-serif", "minWidth": 95},
        ),
    )
    return table, df


def return_footprint(collider, emittance, beam="lhcb1", n_turns=2000):
    """
    Calculates the collider footprint, for beam 1 or beam 2.

    Args:
        collider (Collider): A Collider object.
        emittance (float): The emittance of the beam.
        beam (str, optional): The name of the beam. Defaults to "lhcb1".
        n_turns (int, optional): The number of turns to simulate to compute the footprint. Defaults
            to 2000.

    Returns:
        tuple: A tuple containing the detuning (qx and qy values) of the footprint.
    """
    fp_polar_xm = collider[beam].get_footprint(
        nemitt_x=emittance,
        nemitt_y=emittance,
        n_turns=n_turns,
        linear_rescale_on_knobs=[xt.LinearRescale(knob_name="beambeam_scale", v0=0.0, dv=0.05)],
        freeze_longitudinal=True,
    )

    qx = fp_polar_xm.qx
    qy = fp_polar_xm.qy

    return qx, qy


def compute_knob_str(collider_check):
    whole_str = ""
    l_knobs = []
    with io.StringIO() as buf, redirect_stdout(buf):
        for k in collider_check.collider.vars.keys():
            collider_check.collider.vars[k]._info(limit = None)
            l_knobs.append(k)
            print("****")
        whole_str = buf.getvalue()

    l_knob_str = whole_str.split("****\n")
    d_knob_str = {knob: knob_str for knob, knob_str in zip(l_knobs, l_knob_str)}

    return d_knob_str



"""Thid module contains all the functions used to produce graphs in the app.
"""

# ==================================================================================================
# --- Imports
# ==================================================================================================
# Import from standard library
import logging
from functools import lru_cache

# Import third-party packages
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
import seaborn as sns
from plotly.subplots import make_subplots

# Import local packages
from .resonance import get_working_diagram


# ==================================================================================================
# --- Plotting functions
# ==================================================================================================
def return_radial_background_traces(df_sv):
    """
    Returns a list of 4 scatter traces, representing the (background) radial lines of the collider.
        Useful when plotting the survey.

    Parameters:
    -----------
    df_sv : pandas.DataFrame
        A DataFrame containing the survey data of the collider.

    Returns:
    --------
    list
        A list of plotly.graph_objs.Scattergl objects representing the radial traces.
    """
    # Add 4 radial lines, each parametrized with a different set of x1, x2, y1, y2
    l_traces_background = []
    for x1, x2, y1, y2 in [
        [np.mean(df_sv["X"]), np.mean(df_sv["X"]), -50000, 50000],
        [-50000, 50000, 0, 0],
        [-50000 + np.mean(df_sv["X"]), 50000 + np.mean(df_sv["X"]), -50000, 50000],
        [-50000 + np.mean(df_sv["X"]), 50000 + np.mean(df_sv["X"]), 50000, -50000],
    ]:
        l_traces_background.append(
            go.Scattergl(
                x=[x1, x2],
                y=[y1, y2],
                mode="lines",
                name="Drift space",
                line_color="lightgrey",
                line_width=1,
                hoverinfo="skip",
                showlegend=False,
            )
        )

    # Return result in a list readable by plotly.add_traces()
    return l_traces_background


def return_beam_pipe_trace(df_sv):
    """
    Returns a Plotly trace representing the beam pipe in the survey.

    Args:
        df_sv (pandas.DataFrame): The survey DataFrame containing the X and Z coordinates of the
            beam pipe.

    Returns:
        plotly.graph_objs.Scattergl: The Plotly trace containing the beam coordinates.
    """
    return go.Scattergl(
        x=df_sv["X"],
        y=df_sv["Z"],
        mode="lines",
        name="Drift space",
        line_color="darkgrey",
        line_width=3,
        hoverinfo="skip",
        showlegend=False,
    )


def return_multipole_trace(
    df_elements,
    df_sv,
    order,
    strength_magnification_factor=5000,
    add_ghost_trace=True,
    l_indices_to_keep=None,
    flat=False,
    xaxis=None,
    yaxis=None,
):
    """
    Returns a list of multipole traces (represented as a rectangle).

    Args:
        df_elements (pandas.DataFrame): A DataFrame containing the elements of the accelerator
            lattice.
        df_sv (pandas.DataFrame): DataFrame containing the survey data.
        order (int): Order of the multipole to represent.
        strength_magnification_factor (float, optional): Factor by which to (visually) magnify the
            strength values. Defaults to 5000.
        add_ghost_trace (bool, optional): Whether to add a ghost trace for legend purposes. Defaults
            to True.
        l_indices_to_keep (list, optional): List of element indices to keep (not to represent the
            whole survey). Defaults to None.
        flat (bool, optional): Whether to return a flat (1D) trace. Defaults to False.
        xaxis (dict, optional): Dictionary containing the x-axis range. Defaults to None.
        yaxis (dict, optional): Dictionary containing y-axis range. Defaults to None.

    Returns:
        plotly.graph_objs._scatter.Scatter: A trace representing a multipole of a given order.
    """
    if flat:
        return return_flat_multipole_trace(
            df_elements,
            df_sv,
            order,
            strength_magnification_factor=strength_magnification_factor,
            add_ghost_trace=add_ghost_trace,
            l_indices_to_keep=l_indices_to_keep,
            xaxis=xaxis,
            yaxis=yaxis,
        )
    else:
        return return_circular_multipole_trace(
            df_elements,
            df_sv,
            order,
            strength_magnification_factor=strength_magnification_factor,
            add_ghost_trace=add_ghost_trace,
            l_indices_to_keep=l_indices_to_keep,
        )


def return_circular_multipole_trace(
    df_elements,
    df_sv,
    order,
    strength_magnification_factor=5000,
    add_ghost_trace=True,
    l_indices_to_keep=None,
):
    """
    Returns a list of multipole traces (represented as a rectangle) in a 2D graph.

    Args:
        df_elements (pandas.DataFrame): A DataFrame containing the elements of the accelerator lattice.
        df_sv (pandas.DataFrame): 1 DataFrame containing the survey data.
        order (int): The order of the multipoles to plot (0 for dipoles, 1 for quadrupoles, etc.).
        strength_magnification_factor (float, optional): Factor by which to (visually) magnify the
            strength values. Defaults to 5000.
        add_ghost_trace (bool, optional): Whether to add a ghost trace for legend purposes. Defaults
            to True.
        l_indices_to_keep (list, optional): List of element indices to keep (not to represent the
            whole survey). Defaults to None.

    Returns:
        list: A list of plotly traces representing the multipoles to represent.
    """

    # Get corresponding colors and name for the multipoles
    if order == 0:
        color = px.colors.qualitative.Plotly[0]
        name = "Dipoles"
    elif order == 1:
        color = px.colors.qualitative.Plotly[1]
        name = "Quadrupoles"
    elif order == 2:
        color = px.colors.qualitative.Plotly[-1]
        name = "Sextupoles"
        strength_magnification_factor = strength_magnification_factor / 2
    elif order == 3:
        color = px.colors.qualitative.Plotly[2]
        name = "Octupoles"
        strength_magnification_factor = strength_magnification_factor / 40

    # Function to filter magnet strength
    def return_correct_strength(x):
        try:
            return x[order]
        except:
            return float(x)

    # Get strength of all multipoles of the requested order
    s_knl = df_elements[df_elements._order == order]["knl"].apply(return_correct_strength)

    # Remove zero-strength dipoles and magnify
    s_knl = s_knl[s_knl != 0] * strength_magnification_factor

    # Filter out indices outside of the range if needed
    if l_indices_to_keep is not None:
        s_knl = s_knl[s_knl.index.isin(l_indices_to_keep)]

    # Get corresponding lengths
    s_lengths = df_elements.loc[s_knl.index].length

    # Ghost trace for legend if requested
    if add_ghost_trace:
        ghost_trace = go.Scattergl(
            x=[200000, 200001],
            y=[0, 0],
            mode="lines",
            line=dict(color=color, width=5),
            showlegend=True,
            name=name,
            legendgroup=name,
            # visible="legendonly",
        )

    # Add all multipoles at once, merge them by line width
    dic_trace = {}
    for i, row in df_sv.loc[s_knl.index].iterrows():
        width = (
            np.ceil(s_lengths[i]) if not np.isnan(s_lengths[i]) or np.ceil(s_lengths[i]) == 0 else 1
        )

        if width in dic_trace:
            dic_trace[width]["x"].extend(
                [row["X"], row["X"] + s_knl[i] * np.cos(row["theta"]), None]
            )
            dic_trace[width]["y"].extend(
                [row["Z"], row["Z"] - s_knl[i] * np.sin(row["theta"]), None]
            )
            dic_trace[width]["customdata"].extend([row["name"], row["name"], None])
        else:
            dic_trace[width] = {
                "x": [row["X"], row["X"] + s_knl[i] * np.cos(row["theta"]), None],
                "y": [row["Z"], row["Z"] - s_knl[i] * np.sin(row["theta"]), None],
                "customdata": [row["name"], row["name"], None],
                "mode": "lines",
                "line": dict(
                    color=color,
                    width=width,
                ),
                "showlegend": False,
                "name": row["name"],
                "legendgroup": name,
                "hovertemplate": "Magnet: %{customdata}" + "<extra></extra>",
            }

    l_traces = [go.Scattergl(**dic_trace[width]) for width in dic_trace]

    # Return result in a list readable by plotly.add_traces()
    return [ghost_trace] + l_traces if add_ghost_trace else l_traces


def return_flat_multipole_trace(
    df_elements,
    df_sv,
    order,
    strength_magnification_factor=5000,
    add_ghost_trace=True,
    l_indices_to_keep=None,
    xaxis=None,
    yaxis=None,
):
    """
    Returns a list of multipole traces (each represented as a rectangle) in a 1D graph.

    Args:
        df_elements (pandas.DataFrame): A DataFrame containing the elements of the accelerator
            lattice.
        df_sv (pandas.DataFrame): DataFrame containing the survey data.
        order (int): Order of the multipole to represent.
        strength_magnification_factor (float, optional): Factor by which to (visually) magnify the
            strength values. Defaults to 5000.
        add_ghost_trace (bool, optional): Whether to add a ghost trace for legend purposes. Defaults
            to True.
        l_indices_to_keep (list, optional): List of element indices to keep (not to represent the
            whole survey). Defaults to None.
        xaxis (dict, optional): Dictionary containing the x-axis range. Defaults to None.
        yaxis (dict, optional): Dictionary containing y-axis range. Defaults to None.

    Returns:
        list: List of plotly traces for all multipoles of the given order.
    """

    # Get corresponding colors and name for the multipoles
    if order == 0:
        color = px.colors.qualitative.Plotly[0]
        name = "Dipoles"
        strength_magnification_factor = strength_magnification_factor * 20
    elif order == 1:
        color = px.colors.qualitative.Plotly[1]
        name = "Quadrupoles"
        strength_magnification_factor = strength_magnification_factor * 6
    elif order == 2:
        color = px.colors.qualitative.Plotly[-1]
        name = "Sextupoles"
        strength_magnification_factor = strength_magnification_factor * 2
    elif order == 3:
        color = px.colors.qualitative.Plotly[2]
        name = "Octupoles"
        strength_magnification_factor = strength_magnification_factor / 4

    # Function to filter magnet strength
    def return_correct_strength(x):
        try:
            return x[order]
        except:
            return float(x)

    # Get strength of all multipoles of the requested order
    s_knl = df_elements[df_elements._order == order]["knl"].apply(return_correct_strength)

    # Remove zero-strength dipoles and magnify
    s_knl = s_knl[s_knl != 0] * strength_magnification_factor

    # Filter out indices outside of the range if needed
    if l_indices_to_keep is not None:
        s_knl = s_knl[s_knl.index.isin(l_indices_to_keep)]

    # Get corresponding lengths
    s_lengths = df_elements.loc[s_knl.index].length

    # Ghost trace for legend if requested
    if add_ghost_trace:
        if xaxis is not None and yaxis is not None:
            ghost_trace = go.Scattergl(
                x=[200000, 200001],
                y=[0, 0],
                mode="lines",
                line=dict(color=color, width=5),
                showlegend=True,
                name=name,
                legendgroup=name,
                xaxis=xaxis,
                yaxis=yaxis,
                # visible="legendonly",
            )
        else:
            ghost_trace = go.Scattergl(
                x=[200000, 200001],
                y=[0, 0],
                mode="lines",
                line=dict(color=color, width=5),
                showlegend=True,
                name=name,
                legendgroup=name,
            )

    # Add all multipoles at once, merge them by line width
    dic_trace = {}
    for i, row in df_sv.loc[s_knl.index].iterrows():
        width = (
            np.ceil(s_lengths[i]) if not np.isnan(s_lengths[i]) or np.ceil(s_lengths[i]) == 0 else 1
        )
        if width in dic_trace:
            dic_trace[width]["x"].extend([row["s"], row["s"], None])
            dic_trace[width]["y"].extend([0, s_knl[i], None])
            dic_trace[width]["customdata"].extend([row["name"], row["name"], None])
        else:
            if xaxis is not None and yaxis is not None:
                dic_trace[width] = {
                    "x": [row["s"], row["s"], None],
                    "y": [0, s_knl[i], None],
                    "customdata": [row["name"], row["name"], None],
                    "mode": "lines",
                    "line": dict(
                        color=color,
                        width=width,
                    ),
                    "showlegend": False,
                    "name": row["name"],
                    "legendgroup": name,
                    "hovertemplate": "Magnet: %{customdata}" + "<extra></extra>",
                    "xaxis": xaxis,
                    "yaxis": yaxis,
                }
            else:
                dic_trace[width] = {
                    "x": [row["s"], row["s"], None],
                    "y": [0, s_knl[i], None],
                    "customdata": [row["name"], row["name"], None],
                    "mode": "lines",
                    "line": dict(
                        color=color,
                        width=width,
                    ),
                    "showlegend": False,
                    "name": row["name"],
                    "legendgroup": name,
                    "hovertemplate": "Magnet: %{customdata}" + "<extra></extra>",
                }

    l_traces = [go.Scattergl(**dic_trace[width]) for width in dic_trace]

    # Return result in a list readable by plotly.add_traces()
    return [ghost_trace] + l_traces if add_ghost_trace else l_traces


def return_IP_trace(df_sv, add_ghost_trace=True):
    """
    Returns a list of plotly traces representing the IP elements.

    Parameters:
    -----------
    df_sv : pandas.DataFrame
        The dataframe containing the survey data.
    add_ghost_trace : bool, optional
        Whether to add a ghost trace for the IP elements (for the legend). Default to True.

    Returns:
    --------
    list of plotly.graph_objs._scattergl.Scattergl
        The list of plotly traces representing the IP elements.
    """
    # Get dataframe containing only IP elements
    df_ip = df_sv[df_sv["name"].str.startswith("ip")]

    # Ghost trace for legend if requested
    if add_ghost_trace:
        ghost_trace = go.Scattergl(
            x=[200000, 200000],
            y=[0, 0],
            mode="markers",
            # marker_symbol=218,
            marker_line_color="midnightblue",
            marker_color="grey",
            marker_line_width=2,
            marker_size=15,
            showlegend=True,
            name="IP",
            legendgroup="IP",
            # visible="legendonly",
        )

    # Add all IP at once
    l_traces = [
        go.Scattergl(
            mode="markers",
            x=df_ip["X"],
            y=df_ip["Z"],
            customdata=df_ip["name"],
            # marker_symbol=218,
            marker_line_color="midnightblue",
            marker_color="grey",
            marker_line_width=2,
            marker_size=15,
            name="IP",
            showlegend=False,
            legendgroup="IP",
            hovertemplate="IP: %{customdata}" + "<extra></extra>",
        )
    ]

    # Return result in a list readable by plotly.add_traces()
    return [ghost_trace] + l_traces if add_ghost_trace else l_traces


def return_optic_trace(
    df_sv,
    df_tw,
    type_trace,
    beam_2=False,
    l_indices_to_keep=None,
):
    """
    Returns a plotly scatter trace for a given type of optics data.

    Args:
        df_sv (pandas.DataFrame): DataFrame containing the survey data.
        df_tw (pandas.DataFrame): DataFrame containing the twiss data.
        type_trace (str): Type of trace to plot. Must be one of "betax", "bety", "dx", "dy", "x",
            or "y".
        beam_2 (bool, optional): Whether to plot the trace for beam 2. Defaults to False.
        l_indices_to_keep (list, optional): List of indices to keep, if traces arenot represented
            all along the survey. Defaults to None.

    Returns:
        plotly.graph_objs.Scattergl: Scatter trace for the given type of trace to plot.
    """
    # Get the right twiss dataframe and plotting parameters
    match type_trace:
        case "betax":
            magnification_factor = 1.0
            tw_name = "betx"
            name = r"$\beta_{x2}^{0.8}$" if beam_2 else r"$\beta_{x1}^{0.8}$"
            color = px.colors.qualitative.Plotly[3]
            dash = "dash" if beam_2 else None
            exponent = 0.8

        case "bety":
            magnification_factor = 1.0
            tw_name = "bety"
            name = r"$\beta_{y2}^{0.8}$" if beam_2 else r"$\beta_{y1}^{0.8}$"
            color = px.colors.qualitative.Plotly[4]
            dash = "dash" if beam_2 else None
            exponent = 0.8

        case "dx":
            magnification_factor = 100
            tw_name = "dx"
            name = r"$100D_{x2}$" if beam_2 else r"$100D_{x1}$"
            color = px.colors.qualitative.Plotly[5]
            dash = "dash" if beam_2 else None
            exponent = 1.0

        case "dy":
            magnification_factor = 100
            tw_name = "dy"
            name = r"$100D_{y2}$" if beam_2 else r"$100D_{y1}$"
            color = px.colors.qualitative.Plotly[6]
            dash = "dash" if beam_2 else None
            exponent = 1.0

        case "x":
            magnification_factor = 100000
            tw_name = "x"
            name = r"$10^5x_2$" if beam_2 else r"$10^5x_1$"
            color = px.colors.qualitative.Plotly[7]
            dash = "dash" if beam_2 else None
            exponent = 1.0

        case "y":
            magnification_factor = 100000
            tw_name = "y"
            name = r"$10^5y_2$" if beam_2 else r"$10^5y_1$"
            color = px.colors.qualitative.Plotly[8]
            dash = "dash" if beam_2 else None
            exponent = 1.0

        case _:
            print("The type of trace is not recognized.")

    # Correct for circular projection depending if x-coordinate has been reversed or not
    if beam_2:
        correction = -1
    else:
        correction = 1

    # Only keep requested indices
    if l_indices_to_keep is not None:
        df_sv_temp = df_sv[df_sv.index.isin(l_indices_to_keep)]
        df_tw_temp = df_tw[df_tw.index.isin(l_indices_to_keep)]
    else:
        df_sv_temp = df_sv
        df_tw_temp = df_tw

    # Return the trace
    return go.Scattergl(
        x=[None]
        + list(
            df_sv_temp["X"]
            + df_tw_temp[tw_name] ** exponent
            * correction
            * magnification_factor
            * np.cos(df_sv_temp["theta"])
        )
        + [None],
        y=[None]
        + list(
            df_sv_temp["Z"]
            - df_tw_temp[tw_name] ** exponent * magnification_factor * np.sin(df_sv_temp["theta"])
        )
        + [None],
        mode="lines",
        line=dict(color=color, width=2, dash=dash),
        showlegend=True,
        name=name,
        # visible="legendonly" if hide_optics_traces_initially else True,
        visible=True if (not beam_2 and "bet" in type_trace) else "legendonly",
    )


def add_multipoles_to_fig(
    fig,
    df_elements,
    df_sv,
    l_indices_to_keep,
    add_dipoles,
    add_quadrupoles,
    add_sextupoles,
    add_octupoles,
    flat=False,
    row=None,
    col=None,
    xaxis=None,
    yaxis=None,
):
    """
    Add multipole traces to a given plotly figure.

    Parameters
    ----------
    fig : plotly.graph_objs.Figure
        The figure to which the multipole traces will be added.
    df_elements : pandas.DataFrame
        DataFrame containing the accelerator elements.
    df_sv : pandas.DataFrame
        DataFrame containing the accelerator survey.
    l_indices_to_keep : list of int
        List of indices of the elements to keep in the plot (to only represent part of the survey).
    add_dipoles : bool
        Whether to add dipoles to the plot.
    add_quadrupoles : bool
        Whether to add quadrupoles to the plot.
    add_sextupoles : bool
        Whether to add sextupoles to the plot.
    add_octupoles : bool
        Whether to add octupoles to the plot.
    flat : bool, optional
        Whether to make a 1D or 2D plot (default is False, i.e. 2D plot).
    row : int, optional
        The row index of the subplot to which the traces will be added (default is None).
    col : int, optional
        The column index of the subplot to which the traces will be added (default is None).
    xaxis : str, optional
        The x-axis type of the plot (default is None).
    yaxis : str, optional
        The y-axis type of the plot (default is None).

    Returns
    -------
    plotly.graph_objs.Figure
        The updated figure with the multipole traces added.
    """

    for order, add in zip(
        [0, 1, 2, 3], [add_dipoles, add_quadrupoles, add_sextupoles, add_octupoles]
    ):
        # Add multipole if requested
        if add:
            if row is not None and col is not None:
                l_traces = return_multipole_trace(
                    df_elements,
                    df_sv,
                    order=order,
                    strength_magnification_factor=5000,
                    l_indices_to_keep=l_indices_to_keep,
                    flat=flat,
                    xaxis=xaxis,
                    yaxis=yaxis,
                )
                for trace in l_traces:
                    fig.append_trace(
                        trace,
                        row=row,
                        col=col,
                    )
            else:
                fig.add_traces(
                    return_multipole_trace(
                        df_elements,
                        df_sv,
                        order=order,
                        strength_magnification_factor=5000,
                        l_indices_to_keep=l_indices_to_keep,
                        flat=flat,
                        xaxis=xaxis,
                        yaxis=yaxis,
                    )
                )

    return fig


def add_optics_to_fig(
    fig,
    plot_horizontal_betatron,
    plot_vertical_betatron,
    plot_horizontal_dispersion,
    plot_vertical_dispersion,
    plot_horizontal_position,
    plot_vertical_position,
    df_sv,
    df_tw,
    beam_2=False,
    l_indices_to_keep=None,
):
    """
    Add optics traces to a plotly figure.

    Parameters
    ----------
    fig : plotly.graph_objs._figure.Figure
        The figure to which the traces will be added.
    plot_horizontal_betatron : bool
        Whether to add the horizontal betatron trace.
    plot_vertical_betatron : bool
        Whether to add the vertical betatron trace.
    plot_horizontal_dispersion : bool
        Whether to add the horizontal dispersion trace.
    plot_vertical_dispersion : bool
        Whether to add the vertical dispersion trace.
    plot_horizontal_position : bool
        Whether to add the horizontal position trace.
    plot_vertical_position : bool
        Whether to add the vertical position trace.
    df_sv : pandas.DataFrame
        The survey data.
    df_tw : pandas.DataFrame
        The twiss data.
    beam_2 : bool, optional
        Whether to represent the second beam (default is False).
    l_indices_to_keep : list of int, optional
        The indices of the elements to keep (to only represent part of the survey). Default to None.

    Returns
    -------
    plotly.graph_objs._figure.Figure
        The figure with the added traces.
    """

    # Add horizontal betatron if requested
    if plot_horizontal_betatron:
        fig.add_trace(
            return_optic_trace(
                df_sv,
                df_tw,
                type_trace="betax",
                beam_2=beam_2,
                l_indices_to_keep=l_indices_to_keep,
            )
        )

    # Add vertical betatron if requested
    if plot_vertical_betatron:
        fig.add_trace(
            return_optic_trace(
                df_sv,
                df_tw,
                type_trace="bety",
                beam_2=beam_2,
                l_indices_to_keep=l_indices_to_keep,
            )
        )

    # Add horizontal dispersion if requested
    if plot_horizontal_dispersion:
        fig.add_trace(
            return_optic_trace(
                df_sv,
                df_tw,
                type_trace="dx",
                beam_2=beam_2,
                l_indices_to_keep=l_indices_to_keep,
            )
        )

    # Add vertical dispersion if requested
    if plot_vertical_dispersion:
        fig.add_trace(
            return_optic_trace(
                df_sv,
                df_tw,
                type_trace="dy",
                beam_2=beam_2,
                l_indices_to_keep=l_indices_to_keep,
            )
        )

    # Add horizontal position if requested
    if plot_horizontal_position:
        fig.add_trace(
            return_optic_trace(
                df_sv,
                df_tw,
                type_trace="x",
                beam_2=beam_2,
                l_indices_to_keep=l_indices_to_keep,
            )
        )

    # Add vertical position if requested
    if plot_vertical_position:
        fig.add_trace(
            return_optic_trace(
                df_sv,
                df_tw,
                type_trace="y",
                beam_2=beam_2,
                l_indices_to_keep=l_indices_to_keep,
            )
        )

    return fig


def return_plot_lattice_with_tracking(
    df_sv,
    df_elements,
    df_tw,
    df_sv_2=None,
    df_tw_2=None,
    add_dipoles=True,
    add_quadrupoles=True,
    add_sextupoles=True,
    add_octupoles=True,
    add_IP=True,
    l_indices_to_keep=None,
    plot_horizontal_betatron=True,
    plot_vertical_betatron=True,
    plot_horizontal_dispersion=True,
    plot_vertical_dispersion=True,
    plot_horizontal_position=True,
    plot_vertical_position=True,
    add_optics_beam_2=True,
):
    """
    Returns a Plotly figure showing the lattice of a particle accelerator, with beam tracking data.

    Parameters:
    -----------
    df_sv : pandas.DataFrame
        Dataframe containing the accelerator survey for beam 1.
    df_elements : pandas.DataFrame
        Dataframe containing the accelerator elements.
    df_tw : pandas.DataFrame
        Dataframe containing the twiss parameters of the accelerators for beam 1.
    df_sv_2 : pandas.DataFrame, optional
        Dataframe containing the accelerator survey for beam 2. Default to None.
    df_tw_2 : pandas.DataFrame, optional
        Dataframe containing the twiss parameters of the accelerators for beam 2. Default is None.
    add_dipoles : bool, optional
        Whether to add dipoles to the plot. Default is True.
    add_quadrupoles : bool, optional
        Whether to add quadrupoles to the plot. Default is True.
    add_sextupoles : bool, optional
        Whether to add sextupoles to the plot. Default is True.
    add_octupoles : bool, optional
        Whether to add octupoles to the plot. Default is True.
    add_IP : bool, optional
        Whether to add the interaction point to the plot. Default is True.
    l_indices_to_keep : list of int, optional
        List of indices of the elements to keep in the plot. Default is None (keep all elements).
    plot_horizontal_betatron : bool, optional
        Whether to plot the horizontal betatron function. Default is True.
    plot_vertical_betatron : bool, optional
        Whether to plot the vertical betatron function. Default is True.
    plot_horizontal_dispersion : bool, optional
        Whether to plot the horizontal dispersion. Default is True.
    plot_vertical_dispersion : bool, optional
        Whether to plot the vertical dispersion. Default is True.
    plot_horizontal_position : bool, optional
        Whether to plot the horizontal position. Default is True.
    plot_vertical_position : bool, optional
        Whether to plot the vertical position. Default is True.
    add_optics_beam_2 : bool, optional
        Whether to add optics traces for the second beam. Default is True.

    Returns:
    --------
    fig : plotly.graph_objs._figure.Figure
        The Plotly figure object representing the lattice with beam tracking data.
    """

    # Center X coordinate (otherwise conversion to polar coordinates is not possible)
    X_centered = df_sv["X"] - np.mean(df_sv["X"])

    # Get corresponding angle
    l_theta = np.arctan2(df_sv["Z"], X_centered)

    # Build plotly figure
    fig = go.Figure()

    # Add lines to the bakckground delimit octants
    fig.add_traces(return_radial_background_traces(df_sv))

    # Add beam pipe
    fig.add_trace(return_beam_pipe_trace(df_sv))

    # Add multipoles
    fig = add_multipoles_to_fig(
        fig,
        df_elements,
        df_sv,
        l_indices_to_keep,
        add_dipoles,
        add_quadrupoles,
        add_sextupoles,
        add_octupoles,
        flat=False,
    )

    # Add IP if requested
    if add_IP:
        fig.add_traces(return_IP_trace(df_sv))

    # Add optics traces for beam_1
    fig = add_optics_to_fig(
        fig,
        plot_horizontal_betatron,
        plot_vertical_betatron,
        plot_horizontal_dispersion,
        plot_vertical_dispersion,
        plot_horizontal_position,
        plot_vertical_position,
        df_sv,
        df_tw,
        beam_2=False,
        l_indices_to_keep=l_indices_to_keep,
    )

    # Add optics traces for beam_2 if requested
    if add_optics_beam_2:
        if df_sv_2 is None or df_tw_2 is None:
            print("Warning: df_sv_2 or df_tw_2 is None, beam_2 optics will not be plotted")
        else:
            fig = add_optics_to_fig(
                fig,
                plot_horizontal_betatron,
                plot_vertical_betatron,
                plot_horizontal_dispersion,
                plot_vertical_dispersion,
                plot_horizontal_position,
                plot_vertical_position,
                df_sv_2,
                df_tw_2,
                beam_2=True,
                l_indices_to_keep=l_indices_to_keep,
            )

    # Set general layout for figure
    fig.update_layout(
        # title_text="Survey of the current simulation",
        # title_x=0.5,
        # title_xanchor="center",
        showlegend=True,
        xaxis_range=[df_sv["X"].min() - 300, df_sv["X"].max() + 300],
        yaxis_range=[df_sv["Z"].min() - 300, df_sv["Z"].max() + 300],
        xaxis_showgrid=True,
        xaxis_showticklabels=False,
        yaxis_showgrid=True,
        yaxis_scaleanchor="x",
        yaxis_scaleratio=1,
        yaxis_showticklabels=False,
        legend_tracegroupgap=30,
        # width=1000,
        # height=1000,
        margin=dict(l=10, r=10, b=10, t=10, pad=10),
        dragmode="pan",
    )

    # Make background transparent
    fig.update_layout(
        template="plotly_dark",
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
    )

    return fig


def add_scatter_trace(
    fig,
    x,
    y,
    name,
    row,
    col,
    xaxis,
    yaxis,
    visible=None,
    color=None,
    legendgroup=None,
    dashed=False,
    opacity=0.8,
):
    """
    Add a scatter trace to a plotly figure object.

    Args:
        fig (plotly.graph_objs.Figure): The figure object to add the trace to.
        x (list): The x-axis data for the trace.
        y (list): The y-axis data for the trace.
        name (str): The name of the trace.
        row (int): The row number of the subplot to add the trace to.
        col (int): The column number of the subplot to add the trace to.
        xaxis (str): The x-axis to use for the trace.
        yaxis (str): The y-axis to use for the trace.
        visible (str, optional): The visibility of the trace. Defaults to None.
        color (str, optional): The color of the trace. Defaults to None.
        legendgroup (str, optional): The legend group of the trace. Defaults to None.
        dashed (bool, optional): Whether to use a dashed line for the trace. Defaults to False.
        opacity (float, optional): The opacity of the trace. Defaults to 0.8.

    Returns:
        plotly.graph_objs.Figure: The updated figure object.
    """
    fig.append_trace(
        go.Scattergl(
            x=x,
            y=y,
            mode="lines",
            showlegend=True,
            name=name,
            xaxis=xaxis,
            yaxis=yaxis,
            visible=visible,
            opacity=opacity,
            line=dict(color=color) if not dashed else dict(color=color, dash="dash"),
            # Deactivate legendgroup for now as it doesn't work
            legendgroup=legendgroup,
        ),
        row=row,
        col=col,
    )
    return fig


def return_plot_optics(df_tw_b1, df_tw_b2, df_sv, df_elements, empty=False):
    """
    Returns a Plotly figure object with subplots for magnet traces, beta functions, position
        functions, and dispersion functions.

    Args:
        df_tw_b1 (pandas.DataFrame): DataFrame containing Twiss parameters for beam 1.
        df_tw_b2 (pandas.DataFrame): DataFrame containing Twiss parameters for beam 2.
        df_sv (pandas.DataFrame): DataFrame containing survey data.
        df_elements (pandas.DataFrame): DataFrame containing element data.
        empty (bool, optional): If True, returns an empty figure. Defaults to False.

    Returns:
        plotly.graph_objs._figure.Figure: Plotly figure object with subplots for magnet elements,
            beta functions, position functions, and dispersion functions.
    """
    # Build figure
    fig = make_subplots(rows=4, cols=1, shared_xaxes=True)

    if not empty:
        # Add traces for magnets
        fig = add_multipoles_to_fig(
            fig,
            df_elements,
            df_sv,
            l_indices_to_keep=None,
            add_dipoles=True,
            add_quadrupoles=True,
            add_sextupoles=True,
            add_octupoles=True,
            flat=True,
            row=1,
            col=1,
            xaxis="x",
            yaxis="y",
        )

        # Add traces for beta functions
        fig = add_scatter_trace(
            fig,
            df_tw_b1["s"],
            df_tw_b1["betx"],
            r"$\beta_{x_1}$",
            2,
            1,
            "x",
            "y",
            color="cyan",
            legendgroup="beta-1",
        )
        fig = add_scatter_trace(
            fig,
            df_tw_b1["s"],
            df_tw_b1["bety"],
            r"$\beta_{y_1}$",
            2,
            1,
            "x",
            "y",
            # visible="legendonly",
            color="cyan",
            legendgroup="beta-2",
            dashed=True,
        )
        fig = add_scatter_trace(
            fig,
            df_tw_b2["s"],
            df_tw_b2["betx"],
            r"$\beta_{x_2}$",
            2,
            1,
            "x",
            "y",
            # visible="legendonly",
            color="tomato",
            legendgroup="beta-3",
        )
        fig = add_scatter_trace(
            fig,
            df_tw_b2["s"],
            df_tw_b2["bety"],
            r"$\beta_{y_2}$",
            2,
            1,
            "x",
            "y",
            # visible="legendonly",
            color="tomato",
            legendgroup="beta-4",
            dashed=True,
        )

        # Add traces for position functions
        fig = add_scatter_trace(
            fig,
            df_tw_b1["s"],
            df_tw_b1["x"],
            r"$x_1$",
            3,
            1,
            "x",
            "y2",
            color="cyan",
            legendgroup="position-1",
        )
        fig = add_scatter_trace(
            fig,
            df_tw_b1["s"],
            df_tw_b1["y"],
            r"$y_1$",
            3,
            1,
            "x",
            "y2",
            # visible="legendonly",
            color="cyan",
            legendgroup="position-2",
            dashed=True,
        )
        fig = add_scatter_trace(
            fig,
            df_tw_b2["s"],
            df_tw_b2["x"],
            r"$x_2$",
            3,
            1,
            "x",
            "y2",
            # visible="legendonly",
            color="tomato",
            legendgroup="position-3",
        )
        fig = add_scatter_trace(
            fig,
            df_tw_b2["s"],
            df_tw_b2["y"],
            r"$y_2$",
            3,
            1,
            "x",
            "y2",
            # visible="legendonly",
            color="tomato",
            legendgroup="position-4",
            dashed=True,
        )

        # Add traces for dispersion functions
        fig = add_scatter_trace(
            fig,
            df_tw_b1["s"],
            df_tw_b1["dx"],
            r"$D_{x_1}$",
            4,
            1,
            "x",
            "y3",
            color="cyan",
            legendgroup="dispersion-1",
        )
        fig = add_scatter_trace(
            fig,
            df_tw_b1["s"],
            df_tw_b1["dy"],
            r"$D_{y_1}$",
            4,
            1,
            "x",
            "y3",
            # visible="legendonly",
            color="cyan",
            legendgroup="dispersion-2",
            dashed=True,
        )
        fig = add_scatter_trace(
            fig,
            df_tw_b2["s"],
            df_tw_b2["dx"],
            r"$D_{x_2}$",
            4,
            1,
            "x",
            "y3",
            # visible="legendonly",
            color="tomato",
            legendgroup="dispersion-3",
        )
        fig = add_scatter_trace(
            fig,
            df_tw_b2["s"],
            df_tw_b2["dy"],
            r"$D_{y_2}$",
            4,
            1,
            "x",
            "y3",
            # visible="legendonly",
            color="tomato",
            legendgroup="dispersion-4",
            dashed=True,
        )

        # Add horizontal lines for all ips
        for ip in [1, 2, 5, 8]:
            fig.add_vline(
                x=float(df_tw_b1[df_tw_b1["name"] == "ip" + str(ip)]["s"].iloc[0]),
                line_width=1,
                line_dash="dash",
                line_color="pink",
                annotation_text=f"IP {ip}",
                annotation_position="top right",
            )

    fig.update_layout(
        showlegend=True,
        xaxis_showgrid=True,
        yaxis_showgrid=True,
        legend_tracegroupgap=30,
        dragmode="pan",
        uirevision="Don't change",
        margin=dict(l=10, r=10, b=10, t=10, pad=10),
        template="plotly_dark",
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
    )

    # Update yaxis properties
    fig.update_xaxes(range=[0, df_tw_b1["s"].iloc[-1] + 1])
    fig.update_yaxes(title_text=r"$\beta_{x,y}[m]$", range=[0, 10000], row=2, col=1)
    fig.update_yaxes(title_text=r"(Closed orbit)$_{x,y}[m]$", range=[-0.03, 0.03], row=3, col=1)
    fig.update_yaxes(title_text=r"$D_{x,y}[m]$", range=[-3, 3], row=4, col=1)
    fig.update_xaxes(title_text=r"$s[m]$", row=4, col=1)
    fig.update_yaxes(fixedrange=True)

    return fig


def return_plot_filling_scheme(array_b1, array_b2, i_bunch_b1, i_bunch_b2, beam_beam_schedule):
    """
    Returns a Plotly figure object representing the filling scheme and number of Long-Range/Head-on.

    Parameters:
    -----------
    array_b1 : numpy.ndarray
        Filling scheme for beam 1.
    array_b2 : numpy.ndarray
        Filling scheme for beam 2.
    i_bunch_b1 : int
        Index of the selected bunch for tracking in beam 1.
    i_bunch_b2 : int
        Index of the selected bunch for tracking in beam 2.
    beam_beam_schedule : pandas.DataFrame
        Dataframe containing the beam-beam schedule information.

    Returns:
    --------
    fig : plotly.graph_objs._figure.Figure
        Plotly figure object representing the filling scheme and number beam-beam interactions.
    """
    # ! i_bunch_b2 is not used for now

    # Get indices of slots filled with bunches
    non_zero_indices_b1 = np.nonzero(array_b1)[0]
    non_zero_indices_b2 = np.nonzero(array_b2)[0]

    fig = make_subplots(rows=4, cols=1, shared_xaxes=True)

    # Add the filling scheme for beam 1
    fig.append_trace(
        go.Scattergl(
            x=non_zero_indices_b1,
            y=array_b1[non_zero_indices_b1],
            mode="markers",
            marker=dict(color="cyan", size=5),
            # name="Beam 1",
            xaxis="x",
            yaxis="y",
            showlegend=False,
        ),
        row=1,
        col=1,
    )

    # Add the filling scheme for beam 2
    fig.append_trace(
        go.Scattergl(
            x=non_zero_indices_b2,
            y=array_b2[non_zero_indices_b2] * 2,
            mode="markers",
            marker=dict(color="tomato", size=5),
            # name="Beam 2",
            xaxis="x",
            yaxis="y",
            showlegend=False,
        ),
        row=1,
        col=1,
    )

    # Compute the number of LR in each experiment
    bbs = beam_beam_schedule
    series_collide_atlas_cms = bbs[bbs["collides in ATLAS/CMS"]]["# of LR in ATLAS/CMS"]
    series_not_collide_atlas_cms = bbs[~bbs["collides in ATLAS/CMS"]]["# of LR in ATLAS/CMS"]
    series_collide_lhcb = bbs[bbs["collides in LHCB"]]["# of LR in LHCB"]
    series_not_collide_lhcb = bbs[~bbs["collides in LHCB"]]["# of LR in LHCB"]
    series_collide_alice = bbs[bbs["collides in ALICE"]]["# of LR in ALICE"]
    series_not_collide_alice = bbs[~bbs["collides in ALICE"]]["# of LR in ALICE"]

    # Add the plot for the number of LR in each experiment
    for row, series_collide, series_not_collide in zip(
        [2, 3, 4],
        [series_collide_atlas_cms, series_collide_lhcb, series_collide_alice],
        [
            series_not_collide_atlas_cms,
            series_not_collide_lhcb,
            series_not_collide_alice,
        ],
    ):
        fig.append_trace(
            go.Scattergl(
                x=series_collide.index,
                y=series_collide,
                xaxis="x",
                yaxis="y2",
                mode="markers",
                marker=dict(
                    color="orange",
                    size=5,
                ),
                legendgroup="colliding",
                showlegend=False,
            ),
            row=row,
            col=1,
        )
        fig.append_trace(
            go.Scattergl(
                x=series_not_collide.index,
                y=series_not_collide,
                xaxis="x",
                yaxis="y2",
                mode="markers",
                marker=dict(
                    color="teal",
                    size=5,
                ),
                legendgroup="not_colliding",
                showlegend=False,
            ),
            row=row,
            col=1,
        )

    # Traces for legend
    fig.append_trace(
        go.Scattergl(
            x=series_collide_atlas_cms.index[:1],
            y=series_collide_atlas_cms.values[:1],
            xaxis="x",
            yaxis="y2",
            mode="markers",
            marker=dict(color="orange", size=5),
            legendgroup="colliding",
            name="Colliding in IP",
        ),
        row=2,
        col=1,
    )
    fig.append_trace(
        go.Scattergl(
            x=series_not_collide_atlas_cms.index[:1],
            y=series_not_collide_atlas_cms.values[:1],
            xaxis="x",
            yaxis="y2",
            mode="markers",
            marker=dict(color="teal", size=5),
            legendgroup="not_colliding",
            name="Not colliding in IP",
        ),
        row=2,
        col=1,
    )

    # Add a vertical line (in all subplots) to indicate the bunch selected for tracking
    fig.add_vline(
        x=i_bunch_b1,
        line_width=1,
        line_dash="dash",
        line_color="white",
        annotation_text="Selected bunch",
        annotation_position="top right",
    )

    # Update yaxis properties
    fig.update_xaxes(range=[0, non_zero_indices_b1[-1] + 1])
    fig.update_yaxes(
        title_text=r"Beam",
        range=[0.7, 2.3],
        row=1,
        col=1,
        tickmode="linear",
        tick0=1,
        dtick=1,
        fixedrange=True,
    )
    fig.update_yaxes(title_text=r"#LR in Atlas/CMS", range=[0, 52], row=2, col=1, fixedrange=True)
    fig.update_yaxes(title_text=r"#LR in LHCb", range=[0, 52], row=3, col=1, fixedrange=True)
    fig.update_yaxes(title_text=r"#LR in ALice", range=[0, 52], row=4, col=1, fixedrange=True)
    fig.update_xaxes(title_text=r"25ns slot", row=4, col=1)
    fig.update_yaxes(fixedrange=True)

    # Update layout
    fig.update_layout(
        showlegend=True,
        xaxis_showgrid=True,
        yaxis_showgrid=True,
        dragmode="pan",
        uirevision="Don't change",
        margin=dict(l=20, r=20, b=10, t=30, pad=10),
        template="plotly_dark",
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        legend_x=1,
        legend_y=0.5,
    )
    # fig.update_layout(yaxis=dict(tickmode="linear", tick0=1, dtick=1))

    return fig


def get_indices_of_interest(df_tw, element_1, element_2):
    """
    Return the indices between the two elements provided.

    Args:
        df_tw (pandas.DataFrame): DataFrame containing twiss data.
        element_1 (str): The name of the first element of interest.
        element_2 (str): The name of the second element of interest.

    Returns:
        list: A list of element indices located between element_1 and element_2 (excluded).
    """
    idx_1 = df_tw.loc[df_tw["name"] == element_1].index[0]
    idx_2 = df_tw.loc[df_tw["name"] == element_2].index[0]
    if idx_2 < idx_1:
        return df_tw.loc[: idx_2 + 1].index.union(df_tw.loc[idx_1:].index) 
    return df_tw.loc[idx_1 : idx_2 + 1].index


def return_plot_separation(dic_separation_ip, plane):
    """
    Returns a plotly figure object representing beam-beam separation at the different IPs.

    Args:
        dic_separation_ip (dict): A dictionary containing separation data for the different IPs.
        plane (str): The plane for which separation data is to be plotted. Can be "x", "y", or "xy".

    Returns:
        go.Figure: A plotly figure object representing beam-beam separation at the different IPs.
    """
    fig = make_subplots(
        rows=2,
        cols=2,
        subplot_titles=("IP 1", "IP 2", "IP 5", "IP 8"),
        specs=[
            [{"secondary_y": True}, {"secondary_y": True}],
            [{"secondary_y": True}, {"secondary_y": True}],
        ],
        horizontal_spacing=0.2,
    )
    for idx, n_ip in enumerate([1, 2, 5, 8]):
        s = dic_separation_ip[f"ip{n_ip}"]["s"]
        if plane == "x" or plane == "y":
            sep = np.abs(dic_separation_ip[f"ip{n_ip}"][f"d{plane}_meter"])
            sep_sigma = np.abs(dic_separation_ip[f"ip{n_ip}"][f"d{plane}_sig"])
        elif plane == "xy":
            sep_sigma = np.abs(
                np.sqrt(
                    dic_separation_ip[f"ip{n_ip}"]["dx_sig"] ** 2
                    + dic_separation_ip[f"ip{n_ip}"]["dy_sig"] ** 2
                )
            )

        if plane == "x" or plane == "y":
            # Do the plot
            fig.add_trace(
                go.Scatter(
                    x=s,
                    y=sep,
                    name="Separation at ip " + str(n_ip),
                    legendgroup=" IP " + str(n_ip),
                    mode="lines+markers",
                    line=dict(color="coral", width=1),
                ),
                row=idx // 2 + 1,
                col=idx % 2 + 1,
                secondary_y=False,
            )

        # Plot normalized separation in any case
        fig.add_trace(
            go.Scatter(
                x=s,
                y=sep_sigma,
                name="Normalized separation at ip " + str(n_ip),
                legendgroup=" IP " + str(n_ip),
                mode="lines+markers",
                line=dict(color="cyan", width=1),
            ),
            row=idx // 2 + 1,
            col=idx % 2 + 1,
            secondary_y=True if plane == "y" or plane == "x" else False,
        )

        # fig.add_trace(
        #     go.Scatter(
        #         x=s,
        #         y=[sep] * len(s),
        #         name="Inner normalized separation at ip " + str(n_ip),
        #         legendgroup=" IP " + str(n_ip),
        #         mode="lines+text",
        #         textposition="top left",
        #         line=dict(color="white", width=1, dash="dash"),
        #         text=[""] * (len(s) - 1) + ["Inner normalized separation"],
        #     ),
        #     row=idx // 2 + 1,
        #     col=idx % 2 + 1,
        #     secondary_y=True,
        # )

    for row in range(1, 3):
        for column in range(1, 3):
            fig.update_yaxes(
                title_text=r"$\textrm{B-B separation }[m]$",
                row=row,
                col=column,
                linecolor="coral",
                secondary_y=False,
            )
            fig.update_yaxes(
                title_text=r"$\textrm{B-B separation }[\sigma]$",
                row=row,
                col=column,
                linecolor="cyan",
                secondary_y=True if plane == "y" or plane == "x" else False,
            )
            fig.update_xaxes(title_text=r"$s [m]$", row=row, col=column)

    # fig.update_yaxes(range=[0, 0.25], row = 1, col = 1, secondary_y= False)
    # Use white theme for graph, centered title
    fig.update_layout(
        template="plotly_dark",
        title="Beam-beam separation at the different IPs",
        title_x=0.5,
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        dragmode="pan",
        showlegend=False,
    )

    return fig


def return_plot_separation_3D(dic_position_ip, ip="ip1"):
    """
    Returns a 3D plotly figure showing the beam-beam separation at different interaction points (IPs).

    Parameters:
    -----------
    dic_position_ip : dict
        A dictionary containing the beam position data for different IPs and beams.

    Returns:
    --------
    fig : plotly.graph_objs._figure.Figure
        A plotly figure object containing the 3D beam-beam separation plot.
    """
    logging.info("Starting computation 3D beam-beam separation")

    fig = make_subplots(
        rows=1,
        cols=1,
        # subplot_titles="IP 1",
        specs=[
            [{"type": "scatter3d"}],
        ],
        horizontal_spacing=0.05,
        vertical_spacing=0.05,
    )

    logging.info("Starting ip " + str(ip))
    for beam, color in zip(["lhcb1", "lhcb2"], ["teal", "tomato"]):
        s = dic_position_ip[beam]["tw"][ip]["s"].to_numpy()
        x = dic_position_ip[beam]["tw"][ip]["x"].to_numpy()
        X = dic_position_ip[beam]["sv"][ip]["X"].to_numpy()
        y = dic_position_ip[beam]["tw"][ip]["y"].to_numpy()
        # Y = dic_position_ip[beam]["sv"][ip]["Y"].to_numpy()
        bx = dic_position_ip[beam]["tw"][ip]["betx"].to_numpy()
        by = dic_position_ip[beam]["tw"][ip]["bety"].to_numpy()
        w = np.sqrt((bx + by) / 2)

        if beam == "lhcb2":
            s = s[::-1]
            x = x[::-1]
            X = X[::-1]
            y = y[::-1]
            bx = bx[::-1]
            by = by[::-1]
            w = w[::-1]

        for i in range(0, int(s.shape[0] / 2) - 20, 10):
            fig.add_trace(
                go.Scatter3d(
                    x=s[i - 5 : i + 6],
                    y=x[i - 5 : i + 6] + X[i - 5 : i + 6],
                    z=y[i - 5 : i + 6],
                    mode="lines",
                    line=dict(color=color, width=w[i]),
                    showlegend=False,
                ),
                row=1,
                col=1,
            )
        for i in range(int(s.shape[0] / 2) + 20, s.shape[0], 10):
            fig.add_trace(
                go.Scatter3d(
                    x=s[i - 5 : i + 6],
                    y=x[i - 5 : i + 6] + X[i - 5 : i + 6],
                    z=y[i - 5 : i + 6],
                    mode="lines",
                    line=dict(color=color, width=w[i]),
                    showlegend=False,
                ),
                row=1,
                col=1,
            )
        for i in range(int(s.shape[0] / 2) - 18, int(s.shape[0] / 2) + 19, 1):
            fig.add_trace(
                go.Scatter3d(
                    x=s[i : i + 3],
                    y=x[i : i + 3] + X[i : i + 3],
                    z=y[i : i + 3],
                    mode="lines",
                    line=dict(color=color, width=w[i + 1]),
                    showlegend=False,
                ),
                row=1,
                col=1,
            )
    range_s = (s[0] - 5, s[-1] + 5)
    range_x = range_y = (
        np.min((np.min((x + X)), np.min(y))) - 0.01,
        np.max((np.max((x + X)), np.max(y))) + 0.01,
    )
    fig.update_scenes(
        xaxis=dict(title="s[m]", range=range_s),
        yaxis=dict(title="x[m]", range=range_x),
        zaxis=dict(title="y[m]", range=range_y),
        aspectmode="cube",
        row=1,
        col=1,
    )
    fig.update_layout(
        autosize=False,
        template="plotly_dark",
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        title="3D Beam-beam separation",
        title_x=0.5,
        title_y=0.95,
        margin=dict(l=20, r=20, b=10, t=30),  # , pad=10),
    )
    logging.info("Returning 3D beam-beam separation figure")
    return fig


@lru_cache(maxsize=20)
def compute_footprint_mesh():
    # Get resonance lines first
    traces = get_working_diagram(
        Qx_range=[0.28, 0.35], Qy_range=[0.28, 0.35], order=12, color="white", alpha=0.1
    )
    return traces


def return_plot_footprint(t_array_footprint, qx, qy, title, plot_filtered_web=False):
    """
    Returns a Plotly figure object representing the footprint of a particle beam in the Qx-Qy plane.

    Parameters:
    -----------
    t_array_footprint : tuple of two 1D arrays
        The Qx and Qy coordinates of the footprint mesh.
    qx : float
        The horizontal tune of the beam.
    qy : float
        The vertical tune of the beam.
    title : str
        The title of the plot.
    plot_filtered_web : bool, optional
        Whether to plot the filtered footprint mesh. Default to False.

    Returns:
    --------
    fig : plotly.graph_objs._figure.Figure
        The Plotly figure object representing the footprint plot.
    """
    logging.info("Starting computation footprint figure")
    palette = sns.color_palette("Spectral", 10).as_hex()
    array_qx, array_qy = t_array_footprint
    fig = go.Figure()
    if array_qx.shape[0] > 0:

        # Get resonance lines first
        fig.add_traces(compute_footprint_mesh())

        if plot_filtered_web:
            # Filter the footprint mesh
            for x, y in zip(array_qx, array_qy):
                # Insert additional None when dx or dy is too big
                # to avoid connecting the lines
                x_temp = np.insert(x, np.where(np.abs(np.diff(x)) > 0.003)[0] + 1, None)
                y_temp = np.insert(y, np.where(np.abs(np.diff(x)) > 0.003)[0] + 1, None)
                x_temp = np.insert(x_temp, np.where(np.abs(np.diff(y)) > 0.003)[0] + 1, None)
                y_temp = np.insert(y_temp, np.where(np.abs(np.diff(y)) > 0.003)[0] + 1, None)
                fig.add_trace(
                    go.Scattergl(
                        x=x_temp,
                        y=y_temp,
                        line_color="whitesmoke",
                        opacity=0.3,
                    )
                )
        for idx, (x, y) in enumerate(zip(array_qx.T, array_qy.T)):
            if plot_filtered_web:
                x_temp = np.insert(x, np.where(np.abs(np.diff(x)) > 0.003)[0] + 1, None)
                y_temp = np.insert(y, np.where(np.abs(np.diff(x)) > 0.003)[0] + 1, None)
                x_temp = np.insert(x_temp, np.where(np.abs(np.diff(y)) > 0.003)[0] + 1, None)
                y_temp = np.insert(y_temp, np.where(np.abs(np.diff(y)) > 0.003)[0] + 1, None)
            x_temp = x
            y_temp = y
            fig.add_trace(
                go.Scattergl(
                    x=x_temp,
                    y=y_temp,
                    line_color=palette[9 - idx],
                    mode="markers",
                ),
            )

        fig.add_trace(
            go.Scattergl(
                x=[qx % 1],
                y=[qy % 1],
                mode="markers",
                marker_color="white",
                marker_size=10,
                marker_symbol="x",
            )
        )

        fig.update_layout(
            xaxis=dict(
                range=[
                    np.percentile(array_qx, 10) - 0.001,
                    np.percentile(array_qy, 90) + 0.001,
                ],
            ),
            yaxis=dict(
                range=[
                    np.percentile(array_qx, 10) - 0.003,
                    np.percentile(array_qy, 90) + 0.003,
                ],
            ),
        )

    fig.update_yaxes(
        scaleanchor="x",
        scaleratio=1,
    )

    fig.update_layout(
        title=title,
        title_x=0.5,
        xaxis_title="Qx",
        yaxis_title="Qy",
        # width=500,
        # height=500,
        showlegend=False,
        margin=dict(l=20, r=20, b=10, t=30, pad=10),
        template="plotly_dark",
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        dragmode="pan",
    )

    logging.info("Returning footprint figure")
    return fig
    logging.info("Returning footprint figure")
    return fig
    return fig



# ==================================================================================================
# --- Imports
# ==================================================================================================

# Import third-party packages
import dash_mantine_components as dmc
from dash import dcc


# ==================================================================================================
# --- Footprint Layout
# ==================================================================================================
def return_footprint_layout():
    """
    Returns the layout for the footprint section of the simulation dashboard.

    Returns:
        footprint_layout : dash.development.base_component.Component
            The layout for the footprint section of the simulation dashboard.

    """
    footprint_layout = (
        dmc.Center(
            dmc.Stack(
                children=[
                    dcc.Loading(
                        children=[
                            dmc.Group(
                                children=[
                                    dcc.Graph(
                                        id="footprint-without-bb-b1",
                                        mathjax=True,
                                        config={
                                            "displayModeBar": False,
                                            "scrollZoom": True,
                                            "responsive": True,
                                            "displaylogo": False,
                                        },
                                        style={
                                            "height": "45vh",
                                            "width": "45%",
                                            "margin": "auto",
                                        },
                                    ),
                                    dcc.Graph(
                                        id="footprint-with-bb-b1",
                                        mathjax=True,
                                        config={
                                            "displayModeBar": False,
                                            "scrollZoom": True,
                                            "responsive": True,
                                            "displaylogo": False,
                                        },
                                        style={
                                            "height": "45vh",
                                            "width": "45%",
                                            "margin": "auto",
                                        },
                                    ),
                                ],
                            ),
                            dmc.Group(
                                children=[
                                    dcc.Graph(
                                        id="footprint-without-bb-b2",
                                        mathjax=True,
                                        config={
                                            "displayModeBar": False,
                                            "scrollZoom": True,
                                            "responsive": True,
                                            "displaylogo": False,
                                        },
                                        style={
                                            "height": "45vh",
                                            "width": "45%",
                                            "margin": "auto",
                                        },
                                    ),
                                    dcc.Graph(
                                        id="footprint-with-bb-b2",
                                        mathjax=True,
                                        config={
                                            "displayModeBar": False,
                                            "scrollZoom": True,
                                            "responsive": True,
                                            "displaylogo": False,
                                        },
                                        style={
                                            "height": "45vh",
                                            "width": "45%",
                                            "margin": "auto",
                                        },
                                    ),
                                ],
                            ),
                        ],
                        type="circle",
                        color="cyan",
                        style={"height": "100%", "width": "100%", "margin": "auto"},
                        parent_style={
                            "height": "100%",
                            "width": "100%",
                            "margin": "auto",
                        },
                    ),
                ],
                style={"width": "100%", "margin": "auto"},
            )
        ),
    )
    return footprint_layout



# ==================================================================================================
# --- Imports
# ==================================================================================================

# Import third-party packages
import dash_mantine_components as dmc
from dash import dcc, html


# ==================================================================================================
# --- Survey layout
# ==================================================================================================
def return_survey_layout():
    """
    Returns the layout for the survey page of the simulation dashboard.

    Returns:
        A Dash HTML div element representing the survey layout.
    """
    survey_layout = html.Div(
        children=[
            dmc.Center(
                dmc.Stack(
                    children=[
                        dmc.Center(
                            children=[
                                dmc.Group(
                                    children=[
                                        dmc.Text("Sectors to display: "),
                                        dmc.ChipGroup(
                                            [
                                                dmc.Chip(
                                                    x,
                                                    value=x,
                                                    variant="outline",
                                                    color="cyan",
                                                )
                                                for x in ["8-2", "2-4", "4-6", "6-8"]
                                            ],
                                            id="chips-ip",
                                            value=["4-6"],
                                            multiple=True,
                                            mb=0,
                                        ),
                                    ],
                                    pt=10,
                                ),
                            ],
                        ),
                        dcc.Loading(
                            children=dcc.Graph(
                                id="LHC-layout",
                                mathjax=True,
                                config={
                                    "displayModeBar": False,
                                    "scrollZoom": True,
                                    "responsive": True,
                                    "displaylogo": False,
                                },
                                style={
                                    "height": "90vh",
                                    "width": "100%",
                                    "margin": "auto",
                                },
                            ),
                            type="circle",
                            color="cyan",
                        ),
                    ],
                    style={"width": "100%", "margin": "auto"},
                )
            ),
            dmc.Drawer(
                title="Element information",
                id="drawer-magnets",
                padding="md",
                transition="rotate-left",
                transitionDuration=20,
                zIndex=10000,
                transitionTimingFunction="ease",
                children=dmc.Card(
                    children=[
                        dmc.Group(
                            [
                                dmc.Text(
                                    id="title-element",
                                    children="Element",
                                    weight=500,
                                ),
                                dmc.Badge(
                                    id="type-element",
                                    children="Dipole",
                                    color="blue",
                                    variant="light",
                                ),
                            ],
                            position="apart",
                            mt="md",
                            mb="xs",
                        ),
                        html.Div(
                            id="text-element",
                            children=[
                                dmc.Text(
                                    id="initial-text",
                                    children=(
                                        "Please click on a multipole or an"
                                        " interaction point to get the"
                                        " corresponding knob information."
                                    ),
                                    size="sm",
                                    color="dimmed",
                                ),
                            ],
                        ),
                    ],
                    withBorder=True,
                    shadow="sm",
                    radius="md",
                    style={"width": "100%"},
                ),
            ),
        ],
        style={"width": "100%", "margin": "auto"},
    )
    return survey_layout
